/*
 *  GuiLoader/C++, a C++ binding to GuiLoader
 *  Copyright (c) 2006-2011 Maxim Udushlivy
 *
 *  Permission is hereby granted, free of charge, to any person
 *  obtaining a copy of this software and associated documentation
 *  files (the "Software"), to deal in the Software without
 *  restriction, including without limitation the rights to use, copy,
 *  modify, merge, publish, distribute, sublicense, and/or sell copies
 *  of the Software, and to permit persons to whom the Software is
 *  furnished to do so, subject to the following conditions:
 *
 *  The above copyright notice and this permission notice shall be
 *  included in all copies or substantial portions of the Software.
 *
 *  The software is provided "as is", without warranty of any kind,
 *  express or implied, including but not limited to the warranties of
 *  merchantability, fitness for a particular purpose and
 *  noninfringement. In no event shall the authors or copyright
 *  holders be liable for any claim, damages or other liability,
 *  whether in an action of contract, tort or otherwise, arising from,
 *  out of or in connection with the software or the use or other
 *  dealings in the software.
 */

#ifndef GUILOADER_CPP_GUILOADER_CPP_H
#define GUILOADER_CPP_GUILOADER_CPP_H

#include <string>
#include <sigc++/sigc++.h>
#include <gtkmm.h>
#include <gtkmm/scalebutton.h>

#ifdef _WIN32
	#ifdef GUILOADER_CPP_LIBRARY
		#define GUILOADER_CPP_API __declspec(dllexport)
	#else
		#define GUILOADER_CPP_API __declspec(dllimport)
	#endif
#else
	#define GUILOADER_CPP_API __attribute__ ((visibility("default")))
#endif

#define GUIXML_DECL              \
  {                              \
    const char data[]=

#define GUIXML_LOAD_CPP(loader)  \
    ;                            \
    loader->load_from_string(std::string(data, sizeof(data)-1)); \
  }

#define GUIXML_(SSS)

namespace Gui {

class LoaderInit;
class Loader;
typedef Glib::RefPtr<Loader> PLoader;

class Signal {

public:
	virtual ~Signal() { }

};

template <class T> class TSignal : public Signal {

public:
	virtual bool connect(Glib::Object * object, const T & slot, bool after) = 0;

};

template <class T> struct TDeduce;

template <class TR, class TO>
struct TDeduce<TR (TO::*)()> {
	typedef sigc::slot<TR> Slot;
	typedef void SlotH;
};

template <class T1, class TR, class TO>
struct TDeduce<TR (TO::*)(T1)> {
	typedef sigc::slot<TR, T1> Slot;
	typedef sigc::slot<TR> SlotH;
};

template <class T1, class T2, class TR, class TO>
struct TDeduce<TR (TO::*)(T1, T2)> {
	typedef sigc::slot<TR, T1, T2> Slot;
	typedef sigc::slot<TR, T1> SlotH;
};

template <class T1, class T2, class T3, class TR, class TO>
struct TDeduce<TR (TO::*)(T1, T2, T3)> {
	typedef sigc::slot<TR, T1, T2, T3> Slot;
	typedef sigc::slot<TR, T1, T2> SlotH;
};

template <class T1, class T2, class T3, class T4, class TR, class TO>
struct TDeduce<TR (TO::*)(T1, T2, T3, T4)> {
	typedef sigc::slot<TR, T1, T2, T3, T4> Slot;
	typedef sigc::slot<TR, T1, T2, T3> SlotH;
};

template <class T1, class T2, class T3, class T4, class T5, class TR, class TO>
struct TDeduce<TR (TO::*)(T1, T2, T3, T4, T5)> {
	typedef sigc::slot<TR, T1, T2, T3, T4, T5> Slot;
	typedef sigc::slot<TR, T1, T2, T3, T4> SlotH;
};

template <class T1, class T2, class T3, class T4, class T5, class T6, class TR, class TO>
struct TDeduce<TR (TO::*)(T1, T2, T3, T4, T5, T6)> {
	typedef sigc::slot<TR, T1, T2, T3, T4, T5, T6> Slot;
	typedef sigc::slot<TR, T1, T2, T3, T4, T5> SlotH;
};

template <class T1, class T2, class T3, class T4, class T5, class T6, class T7, class TR, class TO>
struct TDeduce<TR (TO::*)(T1, T2, T3, T4, T5, T6, T7)> {
	typedef sigc::slot<TR, T1, T2, T3, T4, T5, T6, T7> Slot;
	typedef sigc::slot<TR, T1, T2, T3, T4, T5, T6> SlotH;
};

template <class T>
struct TDeduce {
	typedef typename T::function_type func_type;
	typedef typename TDeduce<func_type>::Slot Slot;
	typedef typename TDeduce<func_type>::SlotH SlotH;
};

class Enter {

public:
	GUILOADER_CPP_API Enter();
	GUILOADER_CPP_API ~Enter();

private:
	Enter(const Enter &);
	Enter & operator = (const Enter &);

};

class Loader : public Glib::Object {

public:
	GUILOADER_CPP_API static PLoader create();

	GUILOADER_CPP_API void set_preview_mode(bool preview);
	GUILOADER_CPP_API void set_take_objects(bool take);
	GUILOADER_CPP_API void load_from_string(const Glib::ustring & data);
	GUILOADER_CPP_API void load_from_file(const std::string & filename);
	GUILOADER_CPP_API void clear();

	template <class S, class T> void connect_signals_s(const std::string & name, const T & slot, int count = 1);
	template <class S, class T> void connect_signals(const std::string & name, const T & slot, int count = 1)
		{ connect_signals_s<S>(name, sigc::hide(slot), count); }
	template <class T> void connect_signals_s(const std::string & name, const T & slot, int count = 1)
		{ connect_signals_s<typename TDeduce<T>::SlotH>(name, slot, count); }
	template <class T> void connect_signals(const std::string & name, const T & slot, int count = 1)
		{ connect_signals<typename TDeduce<T>::Slot>(name, slot, count); }

	template <class T> void bind_object(const std::string & name, Glib::RefPtr<T> * ptr);
	template <class T> void bind_object(const std::string & name, T ** ptr);

	typedef sigc::slot<Gtk::Widget *, std::string, std::string, Glib::ustring> CustomConstructor;

	GUILOADER_CPP_API void add_custom_constructor(const std::string & type, const CustomConstructor & slot);

protected:
	enum { ecConnectFailed, ecConnectBadType, ecBindBadType };

	struct Connector {
		Glib::Object * object;
		Signal * signal;
		bool after;
	};

	typedef std::vector<Connector> Connectors;

	GUILOADER_CPP_API Connectors query_connectors(const std::string & name, int count);
	GUILOADER_CPP_API Glib::Object * bind_object(const std::string & name);
	GUILOADER_CPP_API void error(int code);

private:
	Loader();

	Loader(const Loader &);
	Loader & operator = (const Loader &);

};

template <class S, class T> void Loader::connect_signals_s(const std::string & name, const T & slot, int count) {
	Connectors connectors= query_connectors(name, count);
	for(Connectors::iterator it= connectors.begin(); it!=connectors.end(); ++it) {
		TSignal<S> * signal= dynamic_cast<TSignal<S> *>(it->signal);
		if(signal) {
			bool ok= signal->connect(it->object, sigc::bind<Glib::Object *>(slot, it->object), it->after);
			if(!ok)
				error(ecConnectFailed);
		} else
			error(ecConnectBadType);
	}
}

template <class T> void Loader::bind_object(const std::string & name, Glib::RefPtr<T> * ptr) {
	T * result;
	bind_object(name, &result);
	result->reference();
	*ptr= Glib::RefPtr<T>(result);
}

template <class T> void Loader::bind_object(const std::string & name, T ** ptr) {
	T * result= dynamic_cast<T *>(bind_object(name));
	if(!result)
		error(ecBindBadType);
	*ptr= result;
}

}

#endif
